#!/usr/bin/env python3

# std
from pathlib import Path

# 3rd
import pandas as pd
import numpy as np

# ours
from ankipandas.util.misc import invert_dict

tables_ours2anki = {
    "revs": "revlog",
    "cards": "cards",
    "notes": "notes"
}
tables_anki2ours = invert_dict(tables_ours2anki)

fields_file = Path(__file__).parent / "data" / "anki_fields.csv"
fields_df = pd.read_csv(fields_file)

our_tables = sorted(list(tables_ours2anki.keys()))
our_columns = {
    table: sorted(list(
        fields_df[np.logical_and(
            fields_df["Table"] == table,
            fields_df["Default"] == True
        )]["Column"].unique()
    ))
    for table in our_tables
}

columns_ours2anki = {
    table: dict(zip(
        fields_df[fields_df["Table"] == table]["Column"],
        fields_df[fields_df["Table"] == table]["AnkiColumn"]
    ))
    for table in our_tables
}


columns_anki2ours = {
    table: invert_dict(columns_ours2anki[table])
    for table in our_tables
}



#!/usr/bin/env python3

""" DataFrame utilities. """

# 3rd
import pandas as pd


def replace_df_inplace(df: pd.DataFrame, df_new: pd.DataFrame) -> None:
    """ Replace dataframe 'in place'.

    Args:
        df: :class:`pandas.DataFrame` to be replaced
        df_new: :class:`pandas.DataFrame` to replace the previous one

    Returns:
        None
    """
    if df.index.any():
        df.drop(df.index, inplace=True)
    for col in df_new.columns:
        df[col] = df_new[col]
    drop_cols = set(df.columns) - set(df_new.columns)
    if drop_cols:
        df.drop(drop_cols, axis=1, inplace=True)



#!/usr/bin/env python3

# std
import sqlite3

# 3rd
import pandas as pd
import pathlib

# ours
import ankipandas.convenience_functions as convenience
import ankipandas.core_functions as core
from ankipandas.util.dataframe import replace_df_inplace
from ankipandas.columns import columns_anki2ours, our_columns
from ankipandas.util.misc import invert_dict


class AnkiDataFrame(pd.DataFrame):
    #: Additional attributes of a :class:`AnkiDataFrame` that a normal
    #: :class:`pandas.DataFrame` does not posess. These will be copied in the
    #: constructor.
    _attributes = ("db", "db_path", "_anki_table")

    def __init__(self, *args, **kwargs):
        """ Initializes a blank :class:`AnkiDataFrame`.

        .. warning::

            It is recommended to directly initialize this class with the notes,
            cards or revs table, using one of the methods
            :meth:`.notes`, :meth:`.cards` or :meth:`.revs` instead!

        Args:
            *args: Internal use only. See arguments of
                :class:`pandas.DataFrame`.
            **kwargs: Internal use only. See arguments of
                :class:`pandas.DataFrame`.
        """
        super().__init__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], AnkiDataFrame):
            self._copy_attrs_from(args[0])

        # IMPORTANT: Make sure to add all attributes to the class variable
        # :attr:`._attributes`. Also all of them have to be initialized as None!
        # (see the code where we copy attributes).

        #: Opened Anki database (:class:`sqlite3.Connection`)
        self.db = None  # type: sqlite3.Connection

        #: Path to Anki database that is opened as :attr:`.db`
        self.db_path = None  # type: pathlib.Path

        #: Type of anki table: 'notes', 'cards' or 'revlog'. This corresponds to
        #: the meaning of the ID row.
        self._anki_table = None  # type: str

        #: Prefix for fields as columns
        self.fields_as_columns_prefix = "fld_"

    @property
    def _constructor(self):
        """ This needs to be overriden so that any DataFrame operations do not
        return a :class:`pandas.DataFrame` but a :class:`AnkiDataFrame`."""
        def __constructor(*args, **kw):
            df = self.__class__(*args, **kw)
            self._copy_attrs_to(df)
            return df
        return __constructor

    def _copy_attrs_to(self, df):
        """ Copy all additional attributes of this class to another instance.
        Also see :attr:`self._attributes`.
        """
        for attr in self._attributes:
            df.__dict__[attr] = getattr(self, attr, None)

    def _copy_attrs_from(self, df):
        """ Copy all additional attributes of this class from another instance.
        Also see :attr:`self._attributes`.
        """
        for attr in self._attributes:
            self.__dict__[attr] = getattr(df, attr, None)

    # Constructors
    # ==========================================================================

    def _load_db(self, path):
        self.db = core.load_db(path)
        self.db_path = path

    def _get_table(self, path, user, table):
        if not path:
            path = self.db_path
        self._load_db(convenience.db_path_input(path, user=user))

        # Note: Conversion of dtypes happens first ==> use original
        # column names!
        # todo: this needs to be saved in columns.py
        if table == "notes":
            dtypes = {"id": str, "mid": str}
        elif table == "cards":
            dtypes = {"id": str, "nid": str, "did": str}
        elif table == "revs":
            dtypes = {"id": str, "cid": str}
        else:
            raise ValueError("Invalid table name: {}.".format(table))

        df = core.get_table(self.db, table)
        df = df.astype(dtypes)  # type: pd.DataFrame
        df.rename(columns=columns_anki2ours[table], inplace=True)

        if table == "notes":
            # Tags as list, rather than string joined by space
            df["ntags"] = \
                df["ntags"].apply(
                    lambda joined: [item for item in joined.split(" ") if item]
                )
            # Fields as list, rather than as string joined by \x1f
            df["nflds"] = df["nflds"].str.split("\x1f")

            # Model field
            df["nmodel"] = df["mid"].map(core.get_model_names(self.db))

        if table == "cards":
            # Deck field
            df["cdeck"] = df["did"].map(core.get_deck_names(self.db))

        drop_columns = set(df.columns) - set(our_columns[table])
        for drop_column in drop_columns:
            df.drop(drop_column, axis=1, inplace=True)

        replace_df_inplace(self, df)
        self._anki_table = table

    @classmethod
    def _table_constructor(cls, path, user, table):
        new = AnkiDataFrame()
        new._get_table(path, user, table)
        return new

    @classmethod
    def notes(cls, path=None, user=None):
        """ Initialize :class:`AnkiDataFrame` with notes table loaded from Anki
        database.

        Args:
            path: (Search) path to database see :func:`.db_path_input` for more
                information.
            user: Anki user name. See :func:`.db_path_input` for more
                information.

        Example:

        .. code-block:: python

            import ankipandas
            notes = ankipandas.AnkiDataFrame.notes()

        """
        return cls._table_constructor(path, user, "notes")

    @classmethod
    def cards(cls, path=None, user=None):
        """ Initialize :class:`AnkiDataFrame` with cards table loaded from Anki
        database.

        Args:
            path: (Search) path to database see :func:`.db_path_input` for more
                information.
            user: Anki user name. See :func:`.db_path_input` for more
                information.

        Example:

        .. code-block:: python

            import ankipandas
            cards = ankipandas.AnkiDataFrame.cards()

        """
        return cls._table_constructor(path, user, "cards")

    @classmethod
    def revs(cls, path=None, user=None):
        """ Initialize :class:`AnkiDataFrame` with review table loaded from Anki
        database.

        Args:
            path: (Search) path to database see :func:`.db_path_input` for more
                information.
            user: Anki user name. See :func:`.db_path_input` for more
                information.

        Example:

        .. code-block:: python

            import ankipandas
            revs = ankipandas.AnkiDataFrame.revs()

        """
        return cls._table_constructor(path, user, "revs")

    # ==========================================================================

    def _invalid_table(self):
        raise ValueError("Invalid table: {}.".format(self._anki_table))

    # IDs
    # ==========================================================================

    # todo: call nidS etc. to avoid clashes with attributes?

    @property
    def nid(self):
        if self._anki_table in ["notes", "cards"]:
            if "nid" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'nid' column. That was not "
                    "a good idea. Cannot get note ID anymore."
                )
            else:
                return self["nid"]
        elif self._anki_table == "revs":
            return self.cid.map(core.cid2nid(self.db))
        else:
            self._invalid_table()

    @property
    def cid(self):
        if self._anki_table in ["cards", "revs"]:
            if "cid" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'cid' column. That was not "
                    "a good idea. Cannot get card ID anymore."
                )
            else:
                return self["cid"]
        elif self._anki_table == "notes":
            raise ValueError(
                "Notes can belong to multiple cards. Therefore it is impossible"
                " to associate a card ID with them."
            )
        else:
            self._invalid_table()

    @property
    def mid(self):
        if self._anki_table in ["notes"]:
            if "nmodel" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'nmodel' column. That was not"
                    " a good idea. Cannot get model ID anymore."
                )
            else:
                return self["nmodel"].map(
                    invert_dict(core.get_model_names(self.db))
                )
        if self._anki_table in ["revs", "cards"]:
            if "nmodel" in self.columns:
                return self["nmodel"].map(
                    invert_dict(core.get_model_names(self.db))
                )
            else:
                return self.nid.map(core.nid2mid(self.db))
        else:
            self._invalid_table()

    @property
    def did(self):
        if self._anki_table == "cards":
            if "cdeck" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'cdeck' column. That was not "
                    "a good idea. Cannot get deck ID anymore."
                )
            return self["cdeck"].map(invert_dict(core.get_deck_names(self.db)))
        elif self._anki_table == "notes":
            raise ValueError(
                "Notes can belong to multiple decks. Therefore it is impossible"
                " to associate a deck ID with them."
            )
        elif self._anki_table == "revs":
            return self.cid.map(core.cid2did(self.db))
        else:
            self._invalid_table()

    # Merge tables
    # ==========================================================================

    # todo: use .nids rather than .nid_columns
    def merge_notes(self, inplace=False, columns=None,
                    drop_columns=None, prepend="n",
                    prepend_clash_only=True):
        """ Merge note table into existing dataframe.

        Args:
            inplace: If False, return new dataframe, else update old one
            columns: Columns to merge
            drop_columns: Columns to ignore when merging
            prepend: Prepend this string to fields from note table
            prepend_clash_only: Only prepend the ``prepend`` string when column
                names would otherwise clash.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        return core.merge_dfs(
            df=self,
            df_add=AnkiDataFrame.notes(self.db_path),
            id_df="nid",
            id_add="nid",
            inplace=inplace,
            prepend=prepend,
            prepend_clash_only=prepend_clash_only,
            columns=columns,
            drop_columns=drop_columns
        )

    # todo: support merging into notes frame
    # todo: use .cids rather than .cid_columns
    def merge_cards(self, inplace=False, columns=None, drop_columns=None,
                    prepend="c", prepend_clash_only=True):
        """
        Merges information from the card table into the current dataframe.

        Args:
            inplace: If False, return new dataframe, else update old one
            columns:  Columns to merge
            drop_columns:  Columns to ignore when merging
            prepend: Prepend this string to fields from card table
            prepend_clash_only: Only prepend the ``prepend`` string when column
                names would otherwise clash.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        return core.merge_dfs(
            df=self,
            df_add=AnkiDataFrame.cards(self.db_path),
            id_df="cid",
            inplace=inplace,
            columns=columns,
            drop_columns=drop_columns,
            id_add="cid",
            prepend=prepend,
            prepend_clash_only=prepend_clash_only
        )


    # Toggle format
    # ==========================================================================

    def fields_as_columns(self, inplace=False):
        """
        In the 'notes' table, the field contents of the notes is contained in
        one column ('flds') by default. With this method, this column can be
        split up into a new column for every field.

        Args:
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`pandas.DataFrame` if inplace==True, else None
        """
        if "nflds" not in self.columns:
            raise ValueError(
                "Could not find fields column 'nflds'."
            )
        # fixme: What if one field column is one that is already in use?
        prefix = self.fields_as_columns_prefix
        if inplace:
            mids = self.mid.unique()
            for mid in mids:
                df_model = self[self.mid == mid]
                fields = pd.DataFrame(df_model["nflds"].tolist())
                field_names = core.get_field_names(self.db)[str(mid)]
                for field in field_names:
                    if prefix + field not in self.columns:
                        self[prefix + field] = ""
                for ifield, field in enumerate(field_names):
                    self.loc[self.mid == mid, [prefix + field]] = \
                        fields[ifield].tolist()
            self.drop("nflds", axis=1, inplace=True)
        else:
            df = self.copy(True)
            df.fields_as_columns(inplace=True)
            return df

    def fields_as_list(self, inplace=False):
        """
        This reverts :meth:`.fields_as_columns`, all columns that represented
        field contents are now merged into one column 'nflds'.

        Args:
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if inplace:
            mids = self.mid.unique()
            to_drop = []
            for mid in mids:
                fields = core.get_field_names(self.db)[str(mid)]
                fields = [
                    self.fields_as_columns_prefix + field for field in fields
                ]
                print(mid, fields)
                print(self[fields])
                print(pd.Series(self[fields].values.tolist()))
                self.loc[self.mid == mid, "nflds"] = \
                    pd.Series(self[fields].values.tolist())
                # Careful: Do not delete the fields here yet, other models
                # might still use them
                to_drop.extend(fields)
            self.drop(to_drop, axis=1, inplace=True)
        else:
            df = self.copy()  # deep?
            df.fields_as_list(
                inplace=True,
            )
            return df

    # Quick access
    # ==========================================================================

    def _check_tag_col(self):
        if "ntags" not in self.columns:
            raise ValueError(
                "Tag column 'ntags' doesn't exist. Perhaps you forgot to merge "
                "the notes into your table?"
            )

    def has_tag(self, tags=None):
        """ Checks whether row has a certain tag ('ntags' column).

        Args:
            tags: String or list thereof. In the latter case, True is returned
                if the row contains any of the specified tags.
                If None (default), True is returned if the row has any tag at
                all.

        Returns:
            Boolean :class:`pd.Series`
        """
        self._check_tag_col()

        if isinstance(tags, str):
            tags = [tags]

        if tags is not None:

            def _has_tag(other):
                return not set(tags).isdisjoint(other)

            return self["ntags"].apply(_has_tag)

        else:
            return self["ntags"].apply(bool)

    def has_tags(self, tags=None):
        """ Checks whether row contains at least the supplied tags.

        Args:
            tags: String or list thereof.
                If None (default), True is returned if the row has any tag at
                all.

        Returns:
            Boolean :class:`pd.Series`
        """
        if tags is None:
            return self.has_tag(None)
        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]
        _has_tags = set(tags).issubset
        return self["ntags"].apply(_has_tags)

    def add_tag(self, tags, inplace=False):
        """ Adds tag ('ntags' column).

        Args:
            tags: String or list thereof.
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if not inplace:
            df = self.copy()  # deep?
            df.add_tag(tags, inplace=True)
            return df

        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]

        if len(tags) == 0:
            return

        def _add_tags(other):
            return other + sorted(list(set(tags) - set(other)))

        self["ntags"] = self["ntags"].apply(_add_tags)

    def remove_tag(self, tags, inplace=False):
        """ Removes tag ('ntags' column).

        Args:
            tags: String or list thereof. If None, all tags are removed.
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if not inplace:
            df = self.copy()  # deep?
            df.remove_tag(tags, inplace=True)
            return df

        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]

        if tags is not None:

            def _remove_tags(other):
                return [tag for tag in other if tag not in tags]

            self["ntags"] = self["ntags"].apply(_remove_tags)

        else:
            self["ntags"] = pd.Series([[]]*len(self))


    # Help
    # ==========================================================================

    def help_cols(self, column='auto', table='all', ankicolumn='all') \
            -> pd.DataFrame:
        """
        Return a pandas dataframe containing descriptions of every field in the
        anki database. The arguments below help to filter it.

        Args:
            column: Name of a field or column (as used by us) or list thereof.
                If 'auto' (default), all columns from the current dataframe will
                be shown.
                If 'all' no filtering based on the table will be performed
            table: Possible values: 'notes', 'cards', 'revlog' or list thereof.
                If 'all' no filtering based on the table will be performed
            ankicolumn: Name of a field or column (as used by Anki) or list
                thereof.
                If 'all' no filtering based on the table will be performed

        Returns:
            Pandas DataFrame with all matches.
        """
        help_path = pathlib.Path(__file__).parent / "data" / "anki_fields.csv"
        df = pd.read_csv(help_path)
        if column == 'auto':
            column = list(self.columns)
        if table is not 'all':
            if isinstance(table, str):
                table = [table]
            df = df[df["Table"].isin(table)]
        if column is not 'all':
            if isinstance(column, str):
                column = [column]
            df = df[df["Column"].isin(column)]
        if ankicolumn is not 'all':
            if isinstance(ankicolumn, str):
                ankicolumn = [ankicolumn]
            df = df[df["AnkiColumn"].isin(ankicolumn)]

        return df

    def help(self):
        # todo
        return ""



#!/usr/bin/env python3

# std
import unittest

# ours
from ankipandas.util.log import log, get_logger, set_log_level


class TestLogging(unittest.TestCase):
    """ Only tests that things run without error. """
    def test_log(self):
        log.info("Test info")
        log.warning("Test warning")

    def test_get_logger(self):
        get_logger().info("Test info")
        get_logger().warning("Test warning")

    def test_set_log_level(self):
        set_log_level("warning")
        set_log_level("WARNING")
        set_log_level(0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# std
import collections
import unittest
from pathlib import Path
import tempfile
from typing import List, Any

# 3rd
from randomfiletree import sample_random_elements, iterative_gaussian_tree

# ours
import ankipandas.convenience_functions as convenience
from ankipandas.columns import *


def touch_file_in_random_folders(basedir, filename: str, n=1) -> List[Path]:
    """ Create files in random folders.

    Args:
        basedir: Starting directory
        filename: Filename of the files to create
        n: Number of files to create

    Returns:
        List of files that were created.
    """
    files = set()
    for d in sample_random_elements(basedir, n_dirs=n, n_files=0,
                                    onfail="ignore")[0]:
        p = Path(d) / filename
        p.touch()
        files.add(p)
    return list(files)


def flatten_list(lst: List[List[Any]]) -> List[Any]:
    """ Takes a list of lists and returns a list of all elements.

    Args:
        lst: List of Lists

    Returns:
        list
    """
    return [item for sublist in lst for item in sublist]


class TestFindDatabase(unittest.TestCase):
    def setUp(self):
        self.dirs = {
            "nothing": tempfile.TemporaryDirectory(),
            "multiple": tempfile.TemporaryDirectory(),
            "perfect": tempfile.TemporaryDirectory()
        }
        for d in self.dirs.values():
            iterative_gaussian_tree(
                d.name,
                repeat=5,
                nfolders=1,
                min_folders=1,
                nfiles=2,
                min_files=1,
            )
        self.dbs = {
            "nothing": [],
            "multiple": touch_file_in_random_folders(
                self.dirs["multiple"].name, "collection.anki2", 10),
            "perfect":  touch_file_in_random_folders(
                self.dirs["perfect"].name, "collection.anki2", 1)
        }
        self.maxDiff = None

    def test_db_path_input_nexist(self):
        with self.assertRaises(FileNotFoundError):
            convenience.db_path_input("/x/y/z")

    def test_db_path_input_multiple(self):
        with self.assertRaises(ValueError):
            convenience.db_path_input(self.dirs["multiple"].name)

    def test_db_path_input_nothing(self):
        with self.assertRaises(ValueError):
            convenience.db_path_input(self.dirs["nothing"].name)

    def test_db_path_input_perfect(self):
        self.assertEqual(
            convenience.db_path_input(self.dirs["perfect"].name),
            self.dbs["perfect"][0]
        )

    def test__find_database(self):
        for d in self.dirs:
            a = sorted(map(str, flatten_list(
                convenience._find_db(
                    self.dirs[d].name, maxdepth=None, break_on_first=False
                ).values()
            )))
            b = sorted(map(str, self.dbs[d]))
            self.assertListEqual(a, b)

    def test__find_database_filename(self):
        # If doesn't exist
        self.assertEqual(
            convenience._find_db(
                Path("abc/myfilename.txt"), filename="myfilename.txt"
            ),
            {}
        )
        tmpdir = tempfile.TemporaryDirectory()
        dir_path = Path(tmpdir.name) / "myfolder"
        file_path = dir_path / "myfilename.txt"
        dir_path.mkdir()
        file_path.touch()
        self.assertEqual(
            convenience._find_db(file_path, filename="myfilename.txt"),
            collections.defaultdict(list, {"myfolder": [file_path]})
        )
        tmpdir.cleanup()

    def test_find_database(self):
        with self.assertRaises(ValueError):
            convenience.find_db(
                self.dirs["nothing"].name, break_on_first=False
            )
        with self.assertRaises(ValueError):
            convenience.find_db(
                self.dirs["multiple"].name, break_on_first=False
            )
            print(self.dbs["multiple"])
        self.assertEqual(
            str(convenience.find_db(
                self.dirs["perfect"].name, break_on_first=False
            )),
            str(self.dbs["perfect"][0])
        )

    def tearDown(self):
        for d in self.dirs.values():
            d.cleanup()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

""" To install this package, change to the directory of this file and run

    pip3 install --user .

(the ``--user`` flag installs the package for your user account only, otherwise
you will need administrator rights).
"""

# std
from distutils.core import setup
# noinspection PyUnresolvedReferences
import setuptools  # see below (1)
from pathlib import Path

# (1) see https://stackoverflow.com/questions/8295644/
# Without this import, install_requires won't work.


keywords = [
    "anki",
    "pandas",
    "dataframe",
]

description = "Load your anki database as a pandas DataFrame with just one " \
              "line of code!"

this_dir = Path(__file__).resolve().parent

packages = setuptools.find_packages()

with (this_dir / "README.rst").open() as fh:
    long_description = fh.read()

with (this_dir / "ankipandas" / "version.txt").open() as vf:
    version = vf.read().strip()

with (this_dir / "requirements.txt").open() as rf:
    requirements = [
        req.strip() for req in rf.readlines()
        if req.strip() and not req.startswith("#")
    ]


setup(
    name='ankipandas',
    version=version,
    packages=packages,
    url="https://github.com/klieret/ankipandas",
    project_urls={
        "Bug Tracker": "https://github.com/klieret/ankipandas/issues",
        "Documentation": "https://ankipandas.readthedocs.io/",
        "Source Code": "https://github.com/klieret/ankipandas/",
    },
    package_data={
        'ankipandas': ['anki_fields.csv', 'data/*'],
    },
    install_requires=requirements,
    license="MIT",
    keywords=keywords,
    description=description,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Topic :: Education",
        "Topic :: Utilities"
    ],
)



#!/usr/bin/env python3

""" Various utilities of this package.

.. warning::

    These utilties are less aimed at end users and might therefore be subject
    to change.

"""

import ankipandas.util.dataframe
import ankipandas.util.log



#!/usr/bin/env python3

# std
import unittest
import shutil
import tempfile

# ours
from ankipandas.core_functions import *
from ankipandas.ankidf import AnkiDataFrame as AnkiDF
from ankipandas.columns import our_columns


class TestCoreFunctionsRead(unittest.TestCase):
    def setUp(self):
        self.db_path = pathlib.Path(__file__).parent / "data" / \
                       "few_basic_cards" / "collection.anki2"
        self.db = load_db(self.db_path)

    def tearDown(self):
        close_db(self.db)

    def test_cards(self):
        cards = AnkiDF.cards(self.db_path)
        self.assertEqual(len(cards), 3)
        self.assertEqual(
            list(sorted(cards.columns)),
            sorted(our_columns["cards"])
        )

    def test_notes(self):
        notes = AnkiDF.notes(self.db_path)
        self.assertEqual(len(notes), 2)
        self.assertEqual(
            list(sorted(notes.columns)),
            sorted(our_columns["notes"])
        )

    def test_get_revs(self):
        revs = AnkiDF.revs(self.db_path)
        # todo assert length
        self.assertEqual(
            list(sorted(revs.columns)),
            sorted(our_columns["revs"])
        )

    def test_get_deck_info(self):
        get_deck_info(self.db)
        # todo

    def test_get_deck_names(self):
        names = get_deck_names(self.db)
        self.assertDictEqual(
            names,
            {"1": "Default"}
        )

    def test_get_model_info(self):
        get_model_info(self.db)
        # todo

    def test_get_model_names(self):
        names = get_model_names(self.db)
        self.assertIn("Basic", names.values())
        self.assertIn("Cloze", names.values())
        self.assertEqual(len(names), 5)

    def test_get_field_names(self):
        fnames = get_field_names(self.db)
        models = get_model_names(self.db)
        fnames = {
            models[mid]: fnames[mid]
            for mid in models
        }
        self.assertEqual(len(fnames), len(get_model_names(self.db)))
        self.assertListEqual(
            fnames["Basic"], ["Front", "Back"]
        )


class TestCoreWrite(unittest.TestCase):
    def setUp(self):
        self.db_read_path = pathlib.Path(__file__).parent / "data" \
            / "few_basic_cards" / "collection.anki2"
        self.db_read = load_db(self.db_read_path)
        self.db_write_dir = tempfile.TemporaryDirectory()
        self.db_write_path = \
            pathlib.Path(self.db_write_dir.name) / "collection.anki2"
        self._reset()

    def _reset(self):
        shutil.copy(str(self.db_read_path), str(self.db_write_path))
        self.db_write = load_db(self.db_write_path)

    def tearDown(self):
        self.db_read.close()
        self.db_write.close()
        self.db_write_dir.cleanup()

    def _check_db_equal(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revlog = get_table(self.db_read, "revs")
        notes2 = get_table(self.db_write, "notes")
        cards2 = get_table(self.db_write, "cards")
        revlog2 = get_table(self.db_write, "revs")
        self.assertListEqual(
            list(notes.values.tolist()), list(notes2.values.tolist())
        )
        self.assertListEqual(
            list(cards.values.tolist()), list(cards2.values.tolist())
        )
        self.assertListEqual(
            list(revlog.values.tolist()), list(revlog2.values.tolist())
        )

    def test_rw_identical(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revlog = get_table(self.db_read, "revs")
        for mode in ["update", "replace", "append"]:
            with self.subTest(mode=mode):
                self._reset()
                set_notes(self.db_write, notes, mode)
                set_cards(self.db_write, cards, mode)
                set_revs(self.db_write, revlog, mode)
                self._check_db_equal()

    def test_update(self):
        notes2 = get_table(self.db_read, "notes")
        notes = get_table(self.db_read, "notes")
        for mode in ["update", "replace", "append"]:
            with self.subTest(mode=mode):
                self._reset()
                notes2.loc[notes2["id"] == 1555579337683, "tags"] = "mytesttag"
                set_notes(self.db_write, notes2, mode)
                if mode == "append":
                    self._check_db_equal()
                else:
                    notes2 = get_table(self.db_write, "notes")
                    chtag = notes2.loc[notes2["id"] == 1555579337683, "tags"]
                    self.assertListEqual(
                        list(chtag.values.tolist()),
                        ["mytesttag"]
                    )
                    unchanged = notes.loc[notes["id"] != 1555579337683, :]
                    unchanged2 = notes2.loc[notes2["id"] != 1555579337683, :]

                    self.assertListEqual(
                        list(unchanged.values.tolist()),
                        list(unchanged2.values.tolist())
                    )


class TestMergeDfs(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame({
            "id_df": [1, 2, 3, 1, 1],
            "clash": ["a", "b", "c", "a", "a"]
        })
        self.df_add = pd.DataFrame({
            "id_add": [1, 2, 3],
            "value": [4, 5, 6],
            "drop": [7, 8, 9],
            "ignore": [10, 11, 12],
            "clash": [1, 1, 1]
        })

    def test_merge_dfs(self):
        df_merged = merge_dfs(
            self.df,
            self.df_add,
            id_df="id_df",
            id_add="id_add",
            prepend="_",
            columns=["value", "drop", "clash"],
            drop_columns=["id_add", "drop"]
        )
        self.assertListEqual(
            sorted(list(df_merged.columns)),
            ["_clash", "clash", "id_df", "value"]
        )
        self.assertListEqual(sorted(list(df_merged["value"])), [4, 4, 4, 5, 6])

    def test_merge_dfs_prepend_all(self):
        df_merged = merge_dfs(
            self.df,
            self.df_add,
            id_df="id_df",
            id_add="id_add",
            prepend="_",
            prepend_clash_only=False
        )
        self.assertListEqual(
            sorted(list(df_merged.columns)),
            ['_clash', '_drop', '_ignore', '_value', 'clash',
             'id_df']
        )

    def test_merge_dfs_inplace(self):
        df = copy.deepcopy(self.df)
        merge_dfs(
            df,
            self.df_add,
            id_df="id_df",
            id_add="id_add",
            inplace=True
        )
        self.assertListEqual(
            sorted(list(df.columns)),
            ['clash_x', 'clash_y', 'drop', 'id_df', 'ignore', 'value']
        )
        self.assertListEqual(sorted(list(df["value"])), [4, 4, 4, 5, 6])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

""" Most of the actual implementation is done with these functions that allow
a finer control over what is happenin than the convenience functions.
"""

# std
import sqlite3
import json
import pathlib
from functools import lru_cache
import copy

# 3rd
import pandas as pd
import numpy as np

# ours
from ankipandas.util.dataframe import replace_df_inplace
from ankipandas.util.log import log
from ankipandas.columns import columns_anki2ours, tables_ours2anki

CACHE_SIZE = 32


# Open/Close db
# ==============================================================================

def load_db(path):
    """
    Load database from path

    Args:
        path: String or pathlib path.

    Returns:
        sqlite connection
    """
    path = pathlib.Path(path)
    if not path.is_file():
        raise FileNotFoundError(
            "Not a file/file not found: {}".format(path)
        )
    return sqlite3.connect(str(path.resolve()))


def close_db(db: sqlite3.Connection) -> None:
    """ Close the database.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        None
    """
    db.close()


# Basic getters
# ==============================================================================

def get_table(db: sqlite3.Connection, table: str) -> pd.DataFrame:
    """ Get raw table from the Anki database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        table: ``cards``, ``notes`` or ``revs``

    Returns:
        :class:`pandas.DataFrame`
    """

    df = pd.read_sql_query(
        "SELECT * FROM {}".format(tables_ours2anki[table]),
        db
    )
    return df


@lru_cache(CACHE_SIZE)
def get_info(db: sqlite3.Connection) -> dict:
    """
    Get all other information from the databse, e.g. information about models,
    decks etc.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary.
    """
    _df = pd.read_sql_query("SELECT * FROM col ", db)
    assert(len(_df) == 1)
    ret = {}
    for col in _df.columns:
        val = _df[col][0]
        if isinstance(val, str):
            ret[col] = json.loads(val)
        else:
            ret[col] = val
    return ret


# Basic Setters
# ==============================================================================

# fixme: Table names changed
# fixme: Need to change data types again
# fixme: need to pop first letters
# fixme: id_column is now dependent
def _set_table(db: sqlite3.Connection, df: pd.DataFrame, table: str,
               mode: str, id_column="id", same_columns=True,
               drop_new_columns=True
               ) -> None:
    """
    Write table back to database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        df: The :class:`pandas.DataFrame` to write
        table: Table to write to: 'notes', 'cards', 'revlog'
        mode: 'update': Update only existing entries, 'append': Only append new
            entries, but do not modify, 'replace': Append, modify and delete
        same_columns: Check that the columns will stay exactly the same
        drop_new_columns: Drop any columns that are in the new dataframe
            but not in the old table.

    Returns:
        None
    """
    df_old = get_table(db, table)
    old_indices = set(df_old[id_column])
    new_indices = set(df[id_column])
    print("old indices", old_indices)
    print("new indices", new_indices)
    if mode == "update":
        indices = set(old_indices)
    elif mode == "append":
        indices = set(new_indices) - set(old_indices)
        if not indices:
            log.warning(
                "Was told to append to table, but there do not seem to be any"
                " new entries. Returning doing nothing."
            )
            return
    elif mode == "replace":
        indices = set(new_indices)
    else:
        raise ValueError("Unknown mode '{}'.".format(mode))
    # Remove everything not indices
    # fixme: Do we need to make a deepcopy?
    df = df[df[id_column].isin(indices)]
    old_cols = list(df_old.columns)
    new_cols = list(df.columns)
    if drop_new_columns:
        df.drop(set(new_cols) - set(old_cols), axis=1, inplace=True)
    if same_columns:
        if not set(df.columns) == set(df_old.columns):
            raise ValueError("Columns do not match: Old: {}, New: {}".format(
                ", ".join(df_old.columns), ", ".join(df.columns)
            ))
    if mode == "append":
        if_exists = "append"
    else:
        if_exists = "replace"
    df.to_sql(table, db, if_exists=if_exists, index=False)


# Trivially derived getters
# ==============================================================================

# todo: Using decorators here causes the function signatures to be messed up
#  with sphinx but oh well.

@lru_cache(CACHE_SIZE)
def get_deck_info(db: sqlite3.Connection):
    """ Get information about decks.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary
    """
    return get_info(db)["decks"]


@lru_cache(CACHE_SIZE)
def get_deck_names(db: sqlite3.Connection):
    """ Mapping of deck IDs (did) to deck names.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping of deck id to deck name
    """
    dinfo = get_deck_info(db)
    return {
        did: dinfo[did]["name"]
        for did in dinfo
    }


@lru_cache(CACHE_SIZE)
def get_model_info(db: sqlite3.Connection):
    """ Get information about models.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary
    """
    return get_info(db)["models"]


@lru_cache(CACHE_SIZE)
def get_model_names(db: sqlite3.Connection):
    """ Mapping of model IDs (mid) to model names.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping of model id to model name
    """
    minfo = get_model_info(db)
    return {
        mid: minfo[mid]["name"]
        for mid in minfo
    }


@lru_cache(CACHE_SIZE)
def get_field_names(db: sqlite3.Connection):
    """ Get names of the fields in the notes

    Args:
        db: Databse (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping of model id to list of field names
    """
    minfo = get_model_info(db)
    return {
        mid: [
            flds["name"] for flds in minfo[mid]["flds"]
        ]
        for mid in minfo
    }


# Trivially derived setters
# ==============================================================================

def set_notes(db: sqlite3.Connection, df: pd.DataFrame, mode: str) -> None:
    """ Write notes table back into database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        df: :class:`pandas.DataFrame`
        mode: 'update': Update only existing entries, 'append': Only append new
            entries, but do not modify, 'replace': Append, modify and delete

    Returns:
        None
    """
    _set_table(db, df, "notes", mode)


def set_cards(db: sqlite3.Connection, df: pd.DataFrame, mode: str):
    """ Write cards table back into database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        df: :class:`pandas.DataFrame`
        mode: 'update': Update only existing entries, 'append': Only append new
            entries, but do not modify, 'replace': Append, modify and delete

    Returns:
        None
    """
    _set_table(db, df, "cards", mode)


def set_revs(db: sqlite3.Connection, df: pd.DataFrame, mode: str):
    """ Write review table back into database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        df: :class:`pandas.DataFrame`
        mode: 'update': Update only existing entries, 'append': Only append new
            entries, but do not modify, 'replace': Append, modify and delete

    Returns:
        None
    """
    _set_table(db, df, "revs", mode)


# Merging information
# ==============================================================================

@lru_cache(CACHE_SIZE)
def cid2nid(db: sqlite3.Connection) -> dict:
    """ Mapping card ID to note ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    cards = get_table(db, "cards")
    return dict(zip(cards["id"].astype(str), cards["nid"].astype(str)))


@lru_cache(CACHE_SIZE)
def cid2did(db: sqlite3.Connection) -> dict:
    """ Mapping card ID to deck ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    cards = get_table(db, "cards")
    return dict(zip(cards["id"].astype(str), cards["did"].astype(str)))


@lru_cache(CACHE_SIZE)
def nid2mid(db: sqlite3.Connection) -> dict:
    """ Mapping note ID to model ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    notes = get_table(db, "notes")
    return dict(zip(notes["id"].astype(str), notes["mid"].astype(str)))


# fixme: This removes items whenever it can't merge!
# todo: move to util
# todo: id_add needs shouldn't have default
def merge_dfs(df: pd.DataFrame, df_add: pd.DataFrame, id_df: str,
              inplace=False, id_add="id", prepend="", replace=False,
              prepend_clash_only=True, columns=None,
              drop_columns=None):
    """
    Merge information from two dataframes.

    Args:
        df: Original :class:`pandas.DataFrame`
        df_add: :class:`pandas.DataFrame` to be merged with original
            :class:`pandas.DataFrame`
        id_df: Column of original dataframe that contains the id along which
            we merge.
        inplace: If False, return new dataframe, else update old one
        id_add: Column of the new dataframe that contains the id along which
            we merge
        prepend: Prepend a string to the column names from the new dataframe
        replace: Replace columns
        prepend_clash_only: Only prepend string to the column names from the
            new dataframe if there is a name clash.
        columns: Keep only these columns
        drop_columns: Drop these columns

    Returns:
        New merged :class:`pandas.DataFrame`
    """
    # Careful: Do not drop the id column until later (else we can't merge)
    # Still, we want to remove as much as possible here, because it's probably
    # better performing
    if columns:
        df_add = df_add.drop(
            set(df_add.columns)-(set(columns) | {id_add}), axis=1
        )
    if drop_columns:
        df_add = df_add.drop(set(drop_columns) - {id_add}, axis=1)
    # Careful: Rename columns after dropping unwanted ones
    if prepend_clash_only:
        col_clash = set(df.columns) & set(df_add.columns)
        rename_dict = {
            col: prepend + col for col in col_clash
        }
    else:
        rename_dict = {
            col: prepend + col for col in df_add.columns
        }
    df_add = df_add.rename(columns=rename_dict)
    # Careful: Might have renamed id_add as well
    if id_add in rename_dict:
        id_add = rename_dict[id_add]

    if replace:
        # Simply remove all potential clashes
        replaced_columns = set(df_add.columns).intersection(set(df.columns))
        df = df.drop(replaced_columns, axis=1)

    df_merge = df.merge(df_add, left_on=id_df, right_on=id_add)
    # Now remove id_add if it was to be removed
    # Careful: 'in' doesn't work with None
    if (columns and id_add not in columns) or \
            (drop_columns and id_add in drop_columns):
        df_merge.drop(id_add, axis=1, inplace=True)

    # todo: make optional
    # Make sure we don't have two ID columns
    new_id_add_col = id_add
    if id_add in rename_dict:
        new_id_add_col = rename_dict[id_add]
    if new_id_add_col in df_merge.columns and id_df != new_id_add_col:
        print("removing", new_id_add_col)
        df_merge.drop(new_id_add_col, axis=1, inplace=True)

    if inplace:
        return replace_df_inplace(df, df_merge)
    else:
        return df_merge



#!/usr/bin/env python3


def invert_dict(dct: dict) -> dict:
    """ Invert dictionary, i.e. reverse keys and values.

    Args:
        dct: Dictionary

    Returns:
        Dictionary with reversed keys and values.

    Raises:
        :class:`ValueError` if values are not unique.
    """
    if not len(set(dct.values())) == len(dct.values()):
        raise ValueError(
            "Dictionary does not seem to be invertible."
        )
    return {
        value: key for key, value in dct.items()
    }



#!/usr/bin/env python3

""" Most of the functionality of the AnkiDataFrames is already tested in
test_core, because that saves to write a lot of duplicate code.

Everything else is tested here.
"""

# std
import pathlib
import unittest
import copy

# ours
from ankipandas.ankidf import AnkiDataFrame as AnkiDF
from ankipandas.columns import our_columns
import ankipandas.core_functions as core


# todo: add more notes to test deck
class TestAnkiDF(unittest.TestCase):
    def setUp(self):
        self.db_path = pathlib.Path(__file__).parent / "data" / \
                       "few_basic_cards" / "collection.anki2"
        self.db = core.load_db(self.db_path)

        # Do not modify this one!
        self.notes = AnkiDF.notes(self.db_path)
        self.cards = AnkiDF.cards(self.db_path)
        self.revs = AnkiDF.revs(self.db_path)
        self.adfs = {
            "notes": self.notes,
            "cards": self.cards,
            "revs": self.revs
        }

    # Test constructors
    # ==========================================================================

    def test_tags(self):
        notes = AnkiDF.notes(self.db_path)
        self.assertListEqual(
            list(notes.query("nid=='1555579337683'")["ntags"].values)[0],
            ['other_test_tag']
        )
        self.assertListEqual(
            list(notes.query("nid=='1555579352896'")["ntags"].values)[0],
            ['some_test_tag']
        )

    def test_merge_note_info(self):
        merged = AnkiDF.cards(self.db_path).merge_notes()
        self.assertListEqual(
            sorted(list(merged.columns)),
            sorted(list(
                set(our_columns["cards"]) | set(our_columns["notes"])
            ))
        )

    # Test properties
    # ==========================================================================

    # Trivial cases
    # --------------------------------------------------------------------------

    def test_nids_notes(self):
        notes = AnkiDF.notes(self.db_path)
        self.assertIn("nid", list(notes.columns))
        self.assertListEqual(
            list(notes["nid"]),
            list(notes.nid)
        )
        self.assertListEqual(
            list(notes["nid"].unique()),
            list(map(str, core.get_table(self.db, "notes")["id"].unique()))
        )
        self.assertEqual(
            len(notes.nid.unique()),
            len(notes.nid)
        )

    def test_cids_cards(self):
        cards = AnkiDF.cards(self.db_path)
        self.assertIn("cid", list(cards.columns))
        self.assertListEqual(
            list(cards["cid"]),
            list(cards.cid)
        )
        self.assertListEqual(
            list(cards["cid"].unique()),
            list(map(str, core.get_table(self.db, "cards")["id"].unique()))
        )
        self.assertEqual(
            len(cards.cid.unique()),
            len(cards.cid)
        )

    def test_rids_revs(self):
        revs = AnkiDF.revs(self.db_path)
        self.assertIn("rid", list(revs.columns))
        self.assertListEqual(
            list(revs["rid"]),
            list(revs.rid)
        )
        self.assertListEqual(
            list(revs["rid"].unique()),
            list(map(str, core.get_table(self.db, "revs")["id"].unique()))
        )
        self.assertEqual(
            len(revs.rid.unique()),
            len(revs.rid)
        )

    # Slightly more elaborate cases
    # --------------------------------------------------------------------------

    def test_nids_cards(self):
        cards = AnkiDF.cards(self.db_path)
        notes = AnkiDF.notes(self.db_path)
        self.assertListEqual(
            sorted(list(cards.nid.unique())),
            sorted(list(notes.nid.unique()))
        )

    def test_nids_revs(self):
        revs = AnkiDF.revs(self.db_path)
        notes = AnkiDF.notes(self.db_path)
        self.assertListEqual(
            sorted(list(revs.nid.unique())),
            sorted(list(notes.nid.unique()))
        )

    def test_nids_nexist(self):
        nexist = AnkiDF()
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            nexist.nid

    def test_cids_revs(self):
        revs = AnkiDF.revs(self.db_path)
        cards = AnkiDF.cards(self.db_path)
        self.assertListEqual(
            sorted(list(revs.cid.unique())),
            sorted(list(cards.cid.unique()))
        )

    def test_cids_notes(self):
        notes = AnkiDF.notes(self.db_path)
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            notes.cid

    # --------------------------------------------------------------------------

    def test_mids(self):
        mids2s = {
            "notes": set(AnkiDF.notes(self.db_path).mid),
            "cards": set(AnkiDF.cards(self.db_path).mid),
            "revs": set(AnkiDF.revs(self.db_path).mid)
        }
        mids = set(core.get_model_names(self.db).keys())
        for table, mids2 in mids2s.items():
            with self.subTest(table=table):
                self.assertTrue(mids2.issubset(mids))

    def test_dids(self):
        did2s = {
            "cards": set(AnkiDF.cards(self.db_path).did),
            "revs": set(AnkiDF.revs(self.db_path).did)
        }
        dids = set(core.get_deck_names(self.db).keys())
        for table, dids2 in did2s.items():
            with self.subTest(table=table):
                self.assertTrue(dids2.issubset(dids))

    # ==========================================================================

    def test_merge_card_info(self):
        merged = AnkiDF.revs(self.db_path).merge_cards()
        self.assertListEqual(
            sorted(list(merged.columns)),
            sorted(list(
                set(our_columns["revs"]) | set(our_columns["cards"])
            ))
        )

    def test_fields_as_columns(self):
        notes = AnkiDF.notes(self.db_path).fields_as_columns()
        cols = our_columns["notes"].copy()
        cols.remove("nflds")
        prefix = notes.fields_as_columns_prefix
        new_cols = [
            prefix + item
            for item in ["Front", "Back"]
        ]
        self.assertEqual(
            sorted(list(notes.columns)),
            sorted(cols + new_cols)
        )
        # fixme: put pack
        # self.assertEqual(
        #     list(notes.query("model=='Basic'")[prefix + "Front"].unique()),
        #     ["Basic: Front"]
        # )

    def test_fields_as_list(self):
        # Add fields as column, remove original 'flds' column, then
        # add it back from the field columns and see if things still check
        # out
        notes = AnkiDF.notes(self.db_path)
        flds = copy.copy(notes["nflds"].values)
        notes.fields_as_columns(inplace=True)
        notes.fields_as_list(inplace=True)
        self.assertEqual(
            list(flds),
            list(notes["nflds"].values)
        )
        self.assertListEqual(
            sorted(list(notes.columns)),
            sorted(our_columns["notes"])
        )

    # Tags
    # ==========================================================================

    def test_remove_tags(self):
        notes = AnkiDF.notes(self.db_path)
        notes = notes.remove_tag(None)
        self.assertEqual(
            list(set(map(tuple, notes["ntags"]))),
            [()]
        )

    def test_add_tags(self):
        notes = AnkiDF.notes(self.db_path).remove_tag(None).add_tag("1145")
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))),
            [('1145', )]
        )
        notes.add_tag("abc", inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))),
            [('1145', "abc")]
        )
        notes.add_tag(["abc", "def"], inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))),
            [('1145', "abc", "def")]
        )
        notes.add_tag([], inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))),
            [('1145', "abc", "def")]
        )

    def test_has_tag(self):
        notes = AnkiDF.notes(self.db_path).remove_tag(None).add_tag("1145")
        self.assertListEqual(
            list(notes.has_tag("1145").unique()),
            [True]
        )
        self.assertListEqual(
            list(notes.has_tag("asdf").unique()),
            [False]
        )
        self.assertListEqual(
            list(notes.has_tag().unique()),
            [True]
        )
        self.assertListEqual(
            list(notes.has_tag(["asdf", "1145"]).unique()),
            [True]
        )

    def test_has_tag_natural(self):
        notes = AnkiDF.notes(self.db_path)
        self.assertListEqual(
            sorted(list(notes.has_tag(["some_test_tag"]).unique())),
            [False, True]
        )

    def test_has_tags(self):
        notes = AnkiDF.notes(self.db_path).remove_tag(None).add_tag("1145")
        self.assertListEqual(
            list(notes.has_tags("1145").unique()),
            [True]
        )
        self.assertListEqual(
            list(notes.has_tags("asdf").unique()),
            [False]
        )
        self.assertListEqual(
            list(notes.has_tags().unique()),
            [True]
        )
        self.assertListEqual(
            list(notes.has_tags(["asdf", "1145"]).unique()),
            [False]
        )
        notes = notes.add_tag("asdf")
        self.assertListEqual(
            list(notes.has_tags(["asdf", "1145"]).unique()),
            [True]
        )

    def test_remove_tag(self):
        notes = AnkiDF.notes(self.db_path).add_tag(["1145", "asdf"])
        notes.remove_tag("1145", inplace=True)
        self.assertListEqual(
            list(notes.has_tag(["1145"]).unique()),
            [False]
        )
        self.assertListEqual(
            list(notes.has_tag(["asdf"]).unique()),
            [True]
        )


    # Help
    # ==========================================================================

    def test_help_cols(self):
        df = self.notes.help_cols()
        self.assertListEqual(
            list(df.columns),
            ["Column", "AnkiColumn", "Table", "Description", "Native", "Default"]
        )

    # fixme
    @unittest.skip
    def test_help_cols_auto(self):
        for table in self.adfs:
            with self.subTest(table=table):
                df = self.adfs[table].help_cols(column="auto")
                self.assertListEqual(
                    sorted(list(set(df["Column"]))),
                    sorted(list(set(self.adfs[table].columns)))
                )
    # todo: expand tests!

    def test_help(self):
        notes = AnkiDF.notes(self.db_path)
        hlp = notes.help()
        # todo


if __name__ == "__main__":
    unittest.main()
